/*
 *	Copyright, Buzztouch
 *
 *	All rights reserved.
 *
 *	Redistribution and use in source and binary forms, with or without modification, are
 *	permitted provided that the following conditions are met:
 *
 *	Redistributions of source code must retain the above copyright notice which includes the
 *	name(s) of the copyright holders. It must also retain this list of conditions and the
 *	following disclaimer.
 *
 *	Redistributions in binary form must reproduce the above copyright notice, this list
 *	of conditions and the following disclaimer in the documentation and/or other materials
 *	provided with the distribution.
 *
 *	The name Buzztouch (or buzztouch) nor the names of its contributors
 *	may be used to endorse or promote products derived from this software without specific
 *	prior written permission.
 *
 *	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 *	ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 *	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 *	IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 *	INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *	NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *	PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *	WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *	ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
 *	OF SUCH DAMAGE.
 */


#import "BT_database.h"
#import "BT_item.h"

@implementation BT_database

//initWithDatabaseFilename...
-(instancetype)initWithDatabaseFilename:(NSString *)dbFilename{
    [BT_debugger showIt:self message:[NSString stringWithFormat:@"initWithDatabaseFilename: %@", dbFilename]];
    self = [super init];
    if(self){
        
        //set the documents directory path...
        NSArray *paths = NSSearchPathForDirectoriesInDomains(NSDocumentDirectory, NSUserDomainMask, YES);
        self.documentsDirectory = [paths objectAtIndex:0];
        
        //set database filename...
        self.databaseFilename = dbFilename;
        
        //copy the database file into the devices file system...
        [self copyDatabaseFromBundle];
        
        //create database tables if they don't exist...
        [self createDefaultTables];
        
    }
    return self;
}


//copyDatabaseFromBundle...
-(void)copyDatabaseFromBundle{
    
    NSString *destinationPath = [self.documentsDirectory stringByAppendingPathComponent:self.databaseFilename];
    if (![[NSFileManager defaultManager] fileExistsAtPath:destinationPath]) {
        NSString *sourcePath = [[[NSBundle mainBundle] resourcePath] stringByAppendingPathComponent:self.databaseFilename];
        NSError *error;
        [[NSFileManager defaultManager] copyItemAtPath:sourcePath toPath:destinationPath error:&error];
        if(error != nil) {
            [BT_debugger showIt:self message:[NSString stringWithFormat:@"copyDatabaseFromBundle: ERROR: %@", [error localizedDescription]]];
        }else{
            [BT_debugger showIt:self message:[NSString stringWithFormat:@"copyDatabaseFromBundle: \"%@\"", self.databaseFilename]];
        }
    }else{
        [BT_debugger showIt:self message:[NSString stringWithFormat:@"copyDatabaseFromBundle: (exists) not copying \"%@\"", self.databaseFilename]];
    }
    
}

//createDefaultTables...
-(void)createDefaultTables{
    [BT_debugger showIt:self message:[NSString stringWithFormat:@"createDefaultTables%@", @""]];
    
    //tbl_btItems (id, guid, createdUNIX, modifiedUNIX, syncedUNIX created automatically in createTableWithName)...
    NSMutableArray *cols = [[NSMutableArray alloc] init];
    [cols addObject:[[NSArray alloc]initWithObjects:@"itemType", @"TEXT", nil]];
    [cols addObject:[[NSArray alloc]initWithObjects:@"itemTypeLabel", @"TEXT", nil]];
    [cols addObject:[[NSArray alloc]initWithObjects:@"nickname", @"TEXT", nil]];
    [cols addObject:[[NSArray alloc]initWithObjects:@"orderIndex", @"INT", nil]];
    [cols addObject:[[NSArray alloc]initWithObjects:@"jsonVars", @"TEXT", nil]];
    [cols addObject:[[NSArray alloc]initWithObjects:@"controlPanelItemType", @"TEXT", nil]];
    [cols addObject:[[NSArray alloc]initWithObjects:@"parentItemGuid", @"TEXT", nil]];
    [cols addObject:[[NSArray alloc]initWithObjects:@"hasChildItems", @"INT", nil]];
    [self createTableWithName:bt_itemsTableName withColumnInfo:cols];
    
}



//openDatabase...
-(void)openDatabase{
    NSString *destinationPath = [self.documentsDirectory stringByAppendingPathComponent:self.databaseFilename];
    if(sqlite3_open([destinationPath UTF8String], &sqlite3Database) != SQLITE_OK) {
        [BT_debugger showIt:self message:[NSString stringWithFormat:@"openDatabase ERROR %s", sqlite3_errmsg(sqlite3Database)]];
    }
}

//closeDatabase...
-(void)closeDatabase{
    if(sqlite3_close(sqlite3Database) != SQLITE_OK){
        [BT_debugger showIt:self message:[NSString stringWithFormat:@"closeDatabase ERROR %s", sqlite3_errmsg(sqlite3Database)]];
    }
}

//getTableList...
-(NSArray *)getTableList{
    [BT_debugger showIt:self message:[NSString stringWithFormat:@"getTablesInfo: %@", @""]];
    
    //open...
    [self openDatabase];
    
    NSMutableArray *tableInfo = [[NSMutableArray alloc] init];
    sqlite3_stmt* statement;
    NSString *query = [NSString stringWithFormat:@"SELECT name FROM sqlite_master WHERE type='table' and name <> 'sqlite_sequence' %@", @""];
    if(sqlite3_prepare_v2(sqlite3Database, [query UTF8String], -1, &statement, nil) == SQLITE_OK){
        while(sqlite3_step(statement) == SQLITE_ROW){
            NSString *value = [NSString stringWithCString:(const char *)sqlite3_column_text(statement, 0) encoding:NSUTF8StringEncoding];
            [tableInfo addObject:value];
        }
        
        //clean up...
        sqlite3_finalize(statement);
        [self closeDatabase];
        
    }else{
        [BT_debugger showIt:self message:[NSString stringWithFormat:@"getTablesInfo ERROR: %s", sqlite3_errmsg(sqlite3Database)]];
    }
    
    //return...
    return tableInfo;
    
}

//getColumnInfoForTable...
-(NSArray*)getColumnInfoForTable:(NSString *)tableName{
    [BT_debugger showIt:self message:[NSString stringWithFormat:@"getColumnInfoForTable: \"%@\"", tableName]];
    
    /*
     Returns an array of NSString objects, one for each table name.
     */
    
    //open database...
    [self openDatabase];
    
    //returns an array of NSMutableDictionary objects holding the colName and the colType..
    NSMutableArray *ret = [[NSMutableArray alloc] init];
    
    //array of column names and dataTypes from table...
    sqlite3_stmt *statement;
    const char *query = [[NSString stringWithFormat:@"pragma table_info('%@')", tableName] UTF8String];
    if(sqlite3_prepare_v2(sqlite3Database, query, -1, &statement, nil) == SQLITE_OK){
        while(sqlite3_step(statement) == SQLITE_ROW){
            
            NSString *colName = [NSString stringWithUTF8String:(char*)sqlite3_column_text(statement, 1)];
            NSString *colType = [NSString stringWithUTF8String:(char*)sqlite3_column_text(statement, 2)];
            
            NSMutableDictionary *dict = [[NSMutableDictionary alloc] init];
            [dict setObject:colName forKey:@"colName"];
            [dict setObject:colType forKey:@"colType"];
            [ret addObject:dict];
            
        }
        
        //clean up statement...
        sqlite3_finalize(statement);
        [self closeDatabase];
        
    }else{
        [BT_debugger showIt:self message:[NSString stringWithFormat:@"getColumnInfoForTable ERROR getting column info for table \"%@\"", tableName]];
    }
    
    //return...
    return ret;
}

//getIndexInfoForTable...
-(NSArray*)getIndexInfoForTable:(NSString *)tableName{
    [BT_debugger showIt:self message:[NSString stringWithFormat:@"getIndexInfoForTable: %@", tableName]];
    
    /*
     Returns an array of NSDictionaries, each holding index information for
     each index in the table.
     */
    
    //open database...
    [self openDatabase];
    
    //returns an array of NSMutableDictionary objects holding the colName and the colType..
    NSMutableArray *ret = [[NSMutableArray alloc] init];
    
    //array of column names and dataTypes from table...
    sqlite3_stmt *statement;
    const char *query = [[NSString stringWithFormat:@"pragma index_list('%@')", tableName] UTF8String];
    if(sqlite3_prepare_v2(sqlite3Database, query, -1, &statement, nil) == SQLITE_OK){
        while(sqlite3_step(statement) == SQLITE_ROW){
            
            NSString *indexName = [NSString stringWithUTF8String:(char*)sqlite3_column_text(statement, 1)];
            NSString *isUnique = [NSString stringWithUTF8String:(char*)sqlite3_column_text(statement, 2)];
            
            NSMutableDictionary *dict = [[NSMutableDictionary alloc] init];
            [dict setObject:indexName forKey:@"indexName"];
            [dict setObject:isUnique forKey:@"isUnique"];
            [ret addObject:dict];
            
        }
        
        //clean up statement...
        sqlite3_finalize(statement);
        [self closeDatabase];
        
    }else{
        [BT_debugger showIt:self message:[NSString stringWithFormat:@"getIndexInfoForTable ERROR getting column info for table \"%@\"", tableName]];
    }
    
    //return...
    return ret;
}


//createTableWithName...
-(BOOL)createTableWithName:(NSString *)tableName withColumnInfo:(NSArray *)columnInfo{
    
    /*
     creates a database table with provided column information like...
     
     NSArray *column1 = [[NSArray alloc]initWithObjects:@"productName", @"TEXT", nil];
     NSArray *column2 = [[NSArray alloc]initWithObjects:@"quantity", @"INT", nil];
     NSArray *column3 = [[NSArray alloc]initWithObjects:@"price", @"REAL", nil];
     NSArray *column4 = [[NSArray alloc]initWithObjects:@"image", @"BLOB", nil];
     
     Allowed Data Types (called Storage Types in sqllite)...
     TEXT, INT, REAL, BLOB.
     
     For boolean use INT. 0 = true. 1 = false.
     For Dates use TEXT. Sort the results later in your logic layer.
     
     */

    //must provide tableName...
    if([tableName length] < 1){
        [BT_debugger showIt:self theMessage:@"createTableWithName ERROR: no table name provided"];
        return false;
    }
    
    //must provide columnInfo...
    if([columnInfo count] < 1){
        [BT_debugger showIt:self theMessage:@"createTableWithName ERROR: no columnInfo provided"];
        return false;
    }
    
    //return flag...
    BOOL ret = FALSE;
    
    //validate column information...
    for(int i = 0; i < columnInfo.count; i++){
        NSArray *tmpKey = [columnInfo objectAtIndex:i];
        NSString *tmpKeyName = [tmpKey objectAtIndex:0];
        
        //column name cannot contain spaces...
        NSRange spaceRange = [tmpKeyName rangeOfCharacterFromSet:[NSCharacterSet whitespaceCharacterSet]];
        if(spaceRange.location != NSNotFound){
            [BT_debugger showIt:self theMessage:@"createTableWithName ERROR: column names cannot have spaces"];
            return false;
        }
        //id field created automatically...
        if([tmpKeyName isEqualToString:@"id"]){
            [BT_debugger showIt:self theMessage:@"createTableWithName ERROR: columnInfo name cannot be \"id\" (id field created automatically)"];
            return false;
        }
        //guid field created automatically...
        if([tmpKeyName isEqualToString:@"guid"]){
            [BT_debugger showIt:self theMessage:@"createTableWithName ERROR: columnInfo name cannot be \"guid\" (guid field created automatically)"];
            return false;
        }
        //createdUNIX field created automatically...
        if([tmpKeyName isEqualToString:@"createdUNIX"]){
            [BT_debugger showIt:self theMessage:@"createdUNIX ERROR: columnInfo name cannot be \"createdUNIX\" (createdUNIX field created automatically)"];
            return false;
        }
        //modifiedUNIX field created automatically...
        if([tmpKeyName isEqualToString:@"modifiedUNIX"]){
            [BT_debugger showIt:self theMessage:@"modifiedUNIX ERROR: columnInfo name cannot be \"modifiedUNIX\" (modifiedUNIX field created automatically)"];
            return false;
        }
        //syncedUNIX field created automatically...
        if([tmpKeyName isEqualToString:@"syncedUNIX"]){
            [BT_debugger showIt:self theMessage:@"syncedUNIX ERROR: columnInfo name cannot be \"syncedUNIX\" (syncedUNIX field created automatically)"];
            return false;
        }
      }
    
    //open database...
    [self openDatabase];
    
    //return if table exists...
    NSString *qryExists = [NSString stringWithFormat:@"SELECT name FROM sqlite_master WHERE type='table' AND name='%@';", tableName];
    
    //create table if not exists already...
    sqlite3_stmt *statement;
    if(sqlite3_prepare_v2(sqlite3Database, [qryExists UTF8String], -1, &statement, nil) == SQLITE_OK){
        if(sqlite3_step(statement) == SQLITE_ROW){
            [BT_debugger showIt:self message:[NSString stringWithFormat:@"createTableWithName: \"%@\" NOT CREATING, exists", tableName]];
        
            //clean up and bail...
            sqlite3_finalize(statement);
            [self closeDatabase];
            return FALSE;
        
        }else{
            [BT_debugger showIt:self message:[NSString stringWithFormat:@"createTableWithName: \"%@\" CREATING, does not exist", tableName]];
        }
        
        //clean up statement...
        sqlite3_finalize(statement);
        
    }
    
    //create insert, start with starting default columns...
    NSString *createSQL = [NSString stringWithFormat:@"CREATE TABLE IF NOT EXISTS %@" " (", tableName];
    createSQL = [createSQL stringByAppendingString:[NSString stringWithFormat:@"%@", @"id INTEGER PRIMARY KEY AUTOINCREMENT"]];
    createSQL = [createSQL stringByAppendingString:[NSString stringWithFormat:@"%@", @", guid TEXT UNIQUE NOT NULL"]];
    
    //add provided columns...
    for(int i = 0; i < columnInfo.count; i++) {
        NSArray *tmpKey = [columnInfo objectAtIndex:i];
        createSQL = [createSQL stringByAppendingString:[NSString stringWithFormat:@", %@ %@", [tmpKey objectAtIndex:0], [tmpKey objectAtIndex:1]]];
    }
    
    //add trailing default columns...
    createSQL = [createSQL stringByAppendingString:[NSString stringWithFormat:@"%@", @", createdUNIX INT DEFAULT CURRENT_TIMESTAMP"]];
    createSQL = [createSQL stringByAppendingString:[NSString stringWithFormat:@"%@", @", modifiedUNIX INT DEFAULT CURRENT_TIMESTAMP"]];
    createSQL = [createSQL stringByAppendingString:[NSString stringWithFormat:@"%@", @", syncedUNIX INT DEFAULT 0"]];
    createSQL = [createSQL stringByAppendingString:@");"];
    
    char *errorMsg;
    if(sqlite3_exec(sqlite3Database, [createSQL UTF8String], NULL, NULL, &errorMsg) != SQLITE_OK){
        [BT_debugger showIt:self theMessage:[NSString stringWithFormat:@"createTableWithName: ERROR creating table: %s SQL: %@ ", errorMsg, createSQL]];
    
        //clean up and bail...
        [self closeDatabase];
        return FALSE;
        
    }else{
        [self closeDatabase];
        ret = TRUE;
    }
    
    //create indexes on default columns (after executing create statement)...
    [self executeQuery:[NSString stringWithFormat:@"CREATE INDEX IF NOT EXISTS %@ ON %@ (%@)", @"createdUNIX", tableName, @"createdUNIX"]];
    [self executeQuery:[NSString stringWithFormat:@"CREATE INDEX IF NOT EXISTS %@ ON %@ (%@)", @"modifiedUNIX", tableName, @"modifiedUNIX"]];
    [self executeQuery:[NSString stringWithFormat:@"CREATE INDEX IF NOT EXISTS %@ ON %@ (%@)", @"syncedUNIX", tableName, @"syncedUNIX"]];
    
    
    //return...
    return ret;
    
}


//createIndexInTable...
-(BOOL)createIndexInTable:(NSString *)tableName onColumn:(NSString *)columnName{
    [BT_debugger showIt:self message:[NSString stringWithFormat:@"createIndexInTable: \"%@\" on column: \"%@\"", tableName, columnName]];
    
    /*
     creates an index on a colum in a table...
     */
    
    //open database...
    [self openDatabase];
    
    //return flag...
    BOOL ret = FALSE;
    
    //create index sql...
    NSString *createSQL = [NSString stringWithFormat:@"CREATE INDEX IF NOT EXISTS %@ ON %@ (%@)", columnName, tableName, columnName];
    
    char *errorMsg;
    if(sqlite3_exec(sqlite3Database, [createSQL UTF8String], NULL, NULL, &errorMsg) != SQLITE_OK){
        [BT_debugger showIt:self theMessage:[NSString stringWithFormat:@"createIndexInTable: ERROR: %s", errorMsg]];
    }else{
        ret = TRUE;
    }
    
    //close database...
    [self closeDatabase];
    
    //return...
    return ret;
    
}


//insertRowIntoTable...
-(void)insertRowIntoTable:(NSString *)tableName withRow:(NSDictionary *)row{
    [BT_debugger showIt:self message:[NSString stringWithFormat:@"insertRowIntoTable: \"%@\"", tableName]];
    
    /*
     DO NOT CALL THIS METHOD IN A LOOP TO INSERT LIKE-RECORDS IN VOLUME. USE THE
     insertRowsIntoTable METHOD INSTEAD.
     
     To use this method, and insert ONE RECORD, provide an NSDictionary of column names
     and values for one record such as...
     
     NSMutableDictionary *rowData = [[NSMutableDictionary alloc] init];
     [rowData setObject:@"A UNIQUE ID" forKey:@"guid"];
     [rowData setObject:@"David" forKey:@"firstName"];
     [rowData setObject:@"Book" forKey:@"lastName"];
     
     */
    
    //get column information for this table so we now how to prepare the sqlite statement...
    NSArray *tableColData = [self getColumnInfoForTable:tableName];
    
    //open database...
    [self openDatabase];
    
    //insert...
    NSArray *allKeys = [row allKeys];
    NSString * sql = [NSString stringWithFormat:@"INSERT INTO %@ (", tableName];
    for(int i = 0; i < allKeys.count; i++){
        sql = [sql stringByAppendingString:[NSString stringWithFormat:@"%@, ", [allKeys objectAtIndex:i]]];
    }
    sql = [sql substringToIndex:[sql length] - 2];
    sql = [sql stringByAppendingString:@") values ("];
    for (int i = 0; i < allKeys.count; i++) {
        sql = [sql stringByAppendingString:@"?, "];
    }
    sql = [sql substringToIndex:[sql length] - 2];
    sql = [sql stringByAppendingString:@")"];
    
    //bind values...
    sqlite3_stmt *statement;
    const char *query = [sql UTF8String];
    if(sqlite3_prepare_v2(sqlite3Database, query, -1, &statement, nil) == SQLITE_OK){
        for(int i = 0; i < allKeys.count; i++){
            
            //get column information for this field to then prepare the sqlite statement...
            NSString *colType = @"";
            for(NSMutableDictionary *colInfo in tableColData){
                if([colInfo objectForKey:@"colName"] != nil && [colInfo objectForKey:@"colType"] != nil){
                    if([[colInfo objectForKey:@"colName"] isEqualToString:[allKeys objectAtIndex:i]]){
                        colType = [colInfo objectForKey:@"colType"];
                        break;
                    }
                }
            }
            
            //bind TEXT...
            if([colType isEqualToString:@"TEXT"]){
                sqlite3_bind_text(statement, (i + 1), [[row valueForKey:[allKeys objectAtIndex:i]] UTF8String], -1, SQLITE_TRANSIENT);
            }
            //bind INT...
            if([colType isEqualToString:@"INT"]){
                sqlite3_bind_int(statement, (i + 1), [[row valueForKey:[allKeys objectAtIndex:i]] intValue]);
            }
            //bind REAL...
            if([colType isEqualToString:@"REAL"]){
                sqlite3_bind_double(statement, (i + 1), [[row valueForKey:[allKeys objectAtIndex:i]] floatValue]);
            }
            //bind BLOB...
            if([colType isEqualToString:@"BLOB"]){
                NSData *tmpData = [row valueForKey:[allKeys objectAtIndex:i]];
                sqlite3_bind_blob(statement, (i + 1), [tmpData bytes], (int)[tmpData length], SQLITE_TRANSIENT);
            }
            
        }//for...
    }else{
        [BT_debugger showIt:self message:[NSString stringWithFormat:@"insertRowIntoTable ERROR (1) %s", sqlite3_errmsg(sqlite3Database)]];
        return;
    }
    
    
    //execute if bind went ok...
    if(sqlite3_errcode(sqlite3Database) < 1){
        if(sqlite3_step(statement) != SQLITE_DONE){
            [BT_debugger showIt:self message:[NSString stringWithFormat:@"insertRowIntoTable ERROR (2) %s", sqlite3_errmsg(sqlite3Database)]];
        }else{
            sqlite3_finalize(statement);
            [self closeDatabase];
        }
    }else{
        [BT_debugger showIt:self message:[NSString stringWithFormat:@"insertRowIntoTable ERROR (3) %s", sqlite3_errmsg(sqlite3Database)]];
    }
    
    
}


//insertRowsIntoTable...
-(void)insertRowsIntoTable:(NSString *)tableName withRows:(NSArray *)rows{
    [BT_debugger showIt:self message:[NSString stringWithFormat:@"insertRowsIntoTable: \"%@\"", tableName]];
    
    /*
     
     DO NOT CALL THIS METHOD TO INSERT ONE RECORD. USE THE
     insertRowIntoTable METHOD INSTEAD.
     
     To use this method, and insert many like-records, provide an NSArray of NSDictionary
     objects. Like-records means all the rows inserted have the same column names in the
     same order. Naturally each row will contain different values...
     
     NSMutableDictionary *row1Data = [[NSMutableDictionary alloc] init];
     [row1Data setObject:@"A UNIQUE ID" forKey:@"guid"];
     [row1Data setObject:@"David" forKey:@"firstName"];
     [row1Data setObject:@"Book" forKey:@"lastName"];
     
     NSMutableDictionary *row2Data = [[NSMutableDictionary alloc] init];
     [row2Data setObject:@"A UNIQUE ID" forKey:@"guid"];
     [row2Data setObject:@"Bill" forKey:@"firstName"];
     [row2Data setObject:@"Jones" forKey:@"lastName"];
     
     */
    
    //must have rows...
    if([rows count] < 1){
        [BT_debugger showIt:self message:[NSString stringWithFormat:@"insertRowsIntoTable ERROR: %@", @"zero rows provided"]];
        return;
    }
    
    //analyze the first row in the provided array to determine the column types for the insert...
    NSDictionary *firstRow = [rows objectAtIndex:0];
    
    //build an list of column data types to be used in the sql statement....
    NSArray *tableColData = [self getColumnInfoForTable:tableName];
    NSMutableDictionary *colDataTypes = [[NSMutableDictionary alloc] init];
    NSArray *allKeys = [firstRow allKeys];
    for(int i = 0; i < allKeys.count; i++){
        
        //get column information for this field...
        for(NSMutableDictionary *colInfo in tableColData){
            if([colInfo objectForKey:@"colName"] != nil && [colInfo objectForKey:@"colType"] != nil){
                if([[colInfo objectForKey:@"colName"] isEqualToString:[allKeys objectAtIndex:i]]){
                    [colDataTypes setObject:[colInfo objectForKey:@"colType"] forKey:[allKeys objectAtIndex:i]];
                    break;
                }
            }
        }
    }
    
    //open database...
    [self openDatabase];
    
    //insert each row, wrap in sqlite transaction to speed things up...
    char* errorMessage;
    sqlite3_exec(sqlite3Database, "BEGIN TRANSACTION", NULL, NULL, &errorMessage);
    sqlite3_stmt *statement;
    
    NSString * sql = [NSString stringWithFormat:@"INSERT INTO %@ (", tableName];
    for(int i = 0; i < allKeys.count; i++){
        sql = [sql stringByAppendingString:[NSString stringWithFormat:@"%@, ", [allKeys objectAtIndex:i]]];
    }
    sql = [sql substringToIndex:[sql length] - 2];
    sql = [sql stringByAppendingString:@") values ("];
    for (int i = 0; i < allKeys.count; i++) {
        sql = [sql stringByAppendingString:@"?, "];
    }
    sql = [sql substringToIndex:[sql length] - 2];
    sql = [sql stringByAppendingString:@")"];
    const char *query = [sql UTF8String];
    
    if(sqlite3_prepare_v2(sqlite3Database, query, -1, &statement, nil) == SQLITE_OK){
        
        //bind each row...
        for(NSDictionary *row in rows){
            
            //bind each column...
            for(int i = 0; i < allKeys.count; i++){
                
                NSString *colType = [colDataTypes objectForKey:[allKeys objectAtIndex:i]];
                
                //bind TEXT...
                if([colType isEqualToString:@"TEXT"]){
                    sqlite3_bind_text(statement, (i + 1), [[row valueForKey:[allKeys objectAtIndex:i]] UTF8String], -1, SQLITE_TRANSIENT);
                }
                //bind INT...
                if([colType isEqualToString:@"INT"]){
                    sqlite3_bind_int(statement, (i + 1), [[row valueForKey:[allKeys objectAtIndex:i]] intValue]);
                }
                //bind REAL...
                if([colType isEqualToString:@"REAL"]){
                    sqlite3_bind_double(statement, (i + 1), [[row valueForKey:[allKeys objectAtIndex:i]] floatValue]);
                }
                //bind BLOB...
                if([colType isEqualToString:@"BLOB"]){
                    NSData *tmpData = [row valueForKey:[allKeys objectAtIndex:i]];
                    sqlite3_bind_blob(statement, (i + 1), [tmpData bytes], (int)[tmpData length], SQLITE_TRANSIENT);
                }
            }
            
            //execute if bind went ok...
            if(sqlite3_step(statement) != SQLITE_DONE){
                [BT_debugger showIt:self message:[NSString stringWithFormat:@"insertRowsIntoTable ERROR: %s", sqlite3_errmsg(sqlite3Database)]];
            }
            
            //reset statement...
            sqlite3_reset(statement);
            
        }//for...
        
        //execute the batch...
        sqlite3_exec(sqlite3Database, "COMMIT TRANSACTION", NULL, NULL, &errorMessage);
        sqlite3_finalize(statement);
        [self closeDatabase];
        
    }else{
        [BT_debugger showIt:self message:[NSString stringWithFormat:@"insertRowsIntoTable ERROR (1) %s", sqlite3_errmsg(sqlite3Database)]];
    }//prepare...
    
}

//updateRowInTable...
-(void)updateRowInTable:(NSString *)tableName withRow:(NSDictionary *)row{
    
    /*
     DO NOT CALL THIS METHOD IN A LOOP TO UPDATE LIKE-RECORDS IN VOLUME. USE THE
     updateRowsInTable METHOD INSTEAD.
     
     To use this method, and update ONE RECORD, provide an NSDictionary of column names
     and values for the record to update. The "guid" value will be used to determine which row to update.
     
     NSMutableDictionary *rowData = [[NSMutableDictionary alloc] init];
     [rowData setObject:@"JFBDF8477SD" forKey:@"guid"];
     [rowData setObject:@"David" forKey:@"firstName"];
     [rowData setObject:@"Book" forKey:@"lastName"];
     
     */
    
    //row must contain guid key...
    if([row objectForKey:@"guid"] == nil){
        [BT_debugger showIt:self message:[NSString stringWithFormat:@"updateRowInTable: ERROR. No guid provided%@", @""]];
        return;
    }else{
        [BT_debugger showIt:self message:[NSString stringWithFormat:@"updateRowInTable: Updating \"%@\" row with guid: \"%@\"", tableName, [row objectForKey:@"guid"]]];
    }
    
    
    //get column information for this table so we now how to prepare the sqlite statement...
    NSArray *tableColData = [self getColumnInfoForTable:tableName];
    
    //open database...
    [self openDatabase];
    
    //create the update string for the paramatized query...
    NSString *sql = [NSString stringWithFormat:@"UPDATE %@ SET ", tableName];
    NSArray *allKeys = [row allKeys];
    for(int i = 0; i < allKeys.count; i++){
        sql = [sql stringByAppendingString:[NSString stringWithFormat:@"%@ = ?, ", [allKeys objectAtIndex:i]]];
    }
    
    //trim trailing comma...
    sql = [sql substringToIndex:[sql length] - 2];
    
    //append where clause...
    sql = [sql stringByAppendingString:[NSString stringWithFormat:@" WHERE guid = '%@'", [row objectForKey:@"guid"]]];
    
    //bind values...
    sqlite3_stmt *statement;
    const char *query = [sql UTF8String];
    if(sqlite3_prepare_v2(sqlite3Database, query, -1, &statement, nil) == SQLITE_OK){
        for(int i = 0; i < allKeys.count; i++){
            
            //get column information for this field to then prepare the sqlite statement...
            NSString *colType = @"";
            for(NSMutableDictionary *colInfo in tableColData){
                if([colInfo objectForKey:@"colName"] != nil && [colInfo objectForKey:@"colType"] != nil){
                    if([[colInfo objectForKey:@"colName"] isEqualToString:[allKeys objectAtIndex:i]]){
                        colType = [colInfo objectForKey:@"colType"];
                        break;
                    }
                }
            }
            
            //bind TEXT...
            if([colType isEqualToString:@"TEXT"]){
                sqlite3_bind_text(statement, (i + 1), [[row valueForKey:[allKeys objectAtIndex:i]] UTF8String], -1, SQLITE_TRANSIENT);
            }
            //bind INT...
            if([colType isEqualToString:@"INT"]){
                sqlite3_bind_int(statement, (i + 1), [[row valueForKey:[allKeys objectAtIndex:i]] intValue]);
            }
            //bind REAL...
            if([colType isEqualToString:@"REAL"]){
                sqlite3_bind_double(statement, (i + 1), [[row valueForKey:[allKeys objectAtIndex:i]] floatValue]);
            }
            //bind BLOB...
            if([colType isEqualToString:@"BLOB"]){
                NSData *tmpData = [row valueForKey:[allKeys objectAtIndex:i]];
                sqlite3_bind_blob(statement, (i + 1), [tmpData bytes], (int)[tmpData length], SQLITE_TRANSIENT);
            }
            
        }//for...
    }else{
        [BT_debugger showIt:self message:[NSString stringWithFormat:@"updateRowInTable ERROR (1) %s", sqlite3_errmsg(sqlite3Database)]];
        return;
    }
    
    //execute if bind went ok...
    if(sqlite3_errcode(sqlite3Database) < 1){
        if(sqlite3_step(statement) != SQLITE_DONE){
            [BT_debugger showIt:self message:[NSString stringWithFormat:@"updateRowInTable ERROR (2) %s", sqlite3_errmsg(sqlite3Database)]];
        }else{
            sqlite3_finalize(statement);
            [self closeDatabase];
        }
    }else{
        [BT_debugger showIt:self message:[NSString stringWithFormat:@"updateRowInTable ERROR (3) %s", sqlite3_errmsg(sqlite3Database)]];
    }
    
    
}


//updateRowsInTable...
-(void)updateRowsInTable:(NSString *)tableName withRows:(NSArray *)rows{
    [BT_debugger showIt:self message:[NSString stringWithFormat:@"updateRowsInTable: \"%@\"", tableName]];
    
    /*
     
     DO NOT CALL THIS METHOD TO UPDATE ONE RECORD. USE THE
     updateRowInTable METHOD INSTEAD.
     
     To use this method, and update many like-records, provide an NSArray of NSDictionary
     objects. Like-records means all the rows updated have the same column names in the
     same order. Naturally each row will contain different values like...
     
     NSMutableDictionary *row1Data = [[NSMutableDictionary alloc] init];
     [row1Data setObject:@"David" forKey:@"firstName"];
     [row1Data setObject:@"Book" forKey:@"lastName"];
     
     NSMutableDictionary *row2Data = [[NSMutableDictionary alloc] init];
     [row2Data setObject:@"Bill" forKey:@"firstName"];
     [row2Data setObject:@"Jones" forKey:@"lastName"];
     
     */
    
    //must have rows...
    if([rows count] < 1){
        [BT_debugger showIt:self message:[NSString stringWithFormat:@"updateRowsInTable ERROR: %@", @"zero rows provided"]];
        return;
    }
    
    //analyze the first row in the provided array to determine the column types for the insert...
    NSDictionary *firstRow = [rows objectAtIndex:0];
    
    //build an list of column data types to be used in the sql statement....
    NSArray *tableColData = [self getColumnInfoForTable:tableName];
    NSMutableDictionary *colDataTypes = [[NSMutableDictionary alloc] init];
    NSArray *allKeys = [firstRow allKeys];
    for(int i = 0; i < allKeys.count; i++){
        
        //get column information for this field...
        for(NSMutableDictionary *colInfo in tableColData){
            if([colInfo objectForKey:@"colName"] != nil && [colInfo objectForKey:@"colType"] != nil){
                if([[colInfo objectForKey:@"colName"] isEqualToString:[allKeys objectAtIndex:i]]){
                    [colDataTypes setObject:[colInfo objectForKey:@"colType"] forKey:[allKeys objectAtIndex:i]];
                    break;
                }
            }
        }
    }
    
    //open database...
    [self openDatabase];
    
    //insert each row, wrap in sqlite transaction to speed things up...
    char* errorMessage;
    sqlite3_exec(sqlite3Database, "BEGIN TRANSACTION", NULL, NULL, &errorMessage);
    sqlite3_stmt *statement;
    
    //create the update string for the paramatized query...
    NSString *sql = [NSString stringWithFormat:@"UPDATE %@ SET ", tableName];
    for(int i = 0; i < allKeys.count; i++){
        sql = [sql stringByAppendingString:[NSString stringWithFormat:@"%@ = ?, ", [allKeys objectAtIndex:i]]];
    }
    
    //trim trailing comma...
    sql = [sql substringToIndex:[sql length] - 2];
    
    //append where clause...
    sql = [sql stringByAppendingString:@" WHERE guid = ?"];
    
    //prepare statement...
    const char *query = [sql UTF8String];
    if(sqlite3_prepare_v2(sqlite3Database, query, -1, &statement, nil) == SQLITE_OK){
        
        //bind each row...
        for(NSDictionary *row in rows){
            
            //row must have a guid...
            if([row objectForKey:@"guid"]){
                
                //bind each column...
                int i = 0;
                for(i = 0; i < allKeys.count; i++){
                    
                    NSString *colType = [colDataTypes objectForKey:[allKeys objectAtIndex:i]];
                    
                    //bind TEXT...
                    if([colType isEqualToString:@"TEXT"]){
                        sqlite3_bind_text(statement, (i + 1), [[row valueForKey:[allKeys objectAtIndex:i]] UTF8String], -1, SQLITE_TRANSIENT);
                    }
                    //bind INT...
                    if([colType isEqualToString:@"INT"]){
                        sqlite3_bind_int(statement, (i + 1), [[row valueForKey:[allKeys objectAtIndex:i]] intValue]);
                    }
                    //bind REAL...
                    if([colType isEqualToString:@"REAL"]){
                        sqlite3_bind_double(statement, (i + 1), [[row valueForKey:[allKeys objectAtIndex:i]] floatValue]);
                    }
                    //bind BLOB...
                    if([colType isEqualToString:@"BLOB"]){
                        NSData *tmpData = [row valueForKey:[allKeys objectAtIndex:i]];
                        sqlite3_bind_blob(statement, (i + 1), [tmpData bytes], (int)[tmpData length], SQLITE_TRANSIENT);
                    }
                }
                
                //bind guid column (last ? in query)...
                sqlite3_bind_text(statement, (i + 1), [[row valueForKey:@"guid"] UTF8String], -1, SQLITE_TRANSIENT);
                
                //execute if bind went ok...
                if(sqlite3_step(statement) != SQLITE_DONE){
                    [BT_debugger showIt:self message:[NSString stringWithFormat:@"updateRowsInTable ERROR: %s", sqlite3_errmsg(sqlite3Database)]];
                }
                
                //reset statement...
                sqlite3_reset(statement);
                
            }//guid != nil...
            
        }//for...
        
        //execute the batch...
        sqlite3_exec(sqlite3Database, "COMMIT TRANSACTION", NULL, NULL, &errorMessage);
        sqlite3_finalize(statement);
        [self closeDatabase];
     
    }else{
        [BT_debugger showIt:self message:[NSString stringWithFormat:@"updateRowsInTable ERROR (1) %s", sqlite3_errmsg(sqlite3Database)]];
    }//prepare...

    
}


//deleteRowFromTable...
-(void)deleteRowFromTable:(NSString *)tableName withGuid:(NSString *)guid{
    [BT_debugger showIt:self message:[NSString stringWithFormat:@"deleteRowFromTable: \"%@\" with guid = \"%@\"", tableName, guid]];
    
    /*
     
     DO NOT CALL THIS METHOD TO DELETE MULTIPLE RECORDS. USE THE
     deleteRowsFromTable METHOD INSTEAD.
     
     To use this method, and delete one rows, provide an NSString guid
     corrosponding to the record you want to delete...
     
     */
    
    //open database...
    [self openDatabase];
    
    //insert each row, wrap in sqlite transaction to speed things up...
    sqlite3_stmt *statement;
    
    //create the update string for the paramatized query...
    NSString *sql = [NSString stringWithFormat:@"DELETE FROM %@ WHERE guid = '%@'", tableName, guid];
    
    //prepare statement...
    const char *query = [sql UTF8String];
    if(sqlite3_prepare_v2(sqlite3Database, query, -1, &statement, nil) == SQLITE_OK){
        if(sqlite3_step(statement) != SQLITE_DONE){
            [BT_debugger showIt:self message:[NSString stringWithFormat:@"deleteRowFromTable ERROR (1) %s", sqlite3_errmsg(sqlite3Database)]];
        }
        
        //reset statement...
        sqlite3_reset(statement);
        
        //finalize...
        sqlite3_finalize(statement);
        [self closeDatabase];
        
    }//prepare...
    
}



//deleteRowsFromTable...
-(void)deleteRowsFromTable:(NSString *)tableName withGuids:(NSArray *)guids{
    [BT_debugger showIt:self message:[NSString stringWithFormat:@"deleteRowsFromTable: \"%@\"", tableName]];
    
    /*
     
     DO NOT CALL THIS METHOD TO DELETE ONE RECORD. USE THE
     deleteRowFromTable METHOD INSTEAD.
     
     To use this method, and delete one to many rows, provide an NSArray of guid's
     corrosponding to the records you want to delete...
     
     */
    
    //open database...
    [self openDatabase];
    
    //for each guid...
    for(NSString *guid in guids){
        
        //create the update string for the paramatized query...
        NSString *sql = [NSString stringWithFormat:@"DELETE FROM %@ WHERE guid = '%@'", tableName, guid];
        
        //prepare statement...
        sqlite3_stmt *statement;
        const char *query = [sql UTF8String];
        if(sqlite3_prepare_v2(sqlite3Database, query, -1, &statement, nil) == SQLITE_OK){
            if(sqlite3_step(statement) != SQLITE_DONE){
                [BT_debugger showIt:self message:[NSString stringWithFormat:@"deleteRowsFromTable ERROR: %s", sqlite3_errmsg(sqlite3Database)]];
            }
            
            //finalize...
            sqlite3_finalize(statement);
            [self closeDatabase];
            
        }//prepare...
        
    }
    
}


//executeQuery...
-(void)executeQuery:(NSString *)sqlQuery{
    [BT_debugger showIt:self message:[NSString stringWithFormat:@"executeQuery: \"%@\"", sqlQuery]];
    
    /*
     Executes a generic sql query such as a delete statement, returns no result.
     It's best NOT to use this generic method to add or update rows. Use
     insertRowIntoTable or updateRowInTable instead.
     */
    
    [self openDatabase];
    char *errorMsg;
    if(sqlite3_exec(sqlite3Database, [sqlQuery UTF8String], NULL, NULL, &errorMsg) != SQLITE_OK){
        [BT_debugger showIt:self message:[NSString stringWithFormat:@"executeQuery ERROR: %s", errorMsg]];
    }
    [self closeDatabase];
    
}



//getOneValue...
-(id)getOneValue:(NSString *)query{
    [BT_debugger showIt:self message:[NSString stringWithFormat:@"getOneValue: \"%@\"", query]];
    
    /*
     Returns an object (string, integer, NSData, etc.) containing the value from one column.
     Query example: SELECT column FROM tbl_name WHERE column = '12345';
     
     NOTE: If query contains SUM, COUNT, AVG, MAX or MIN an aggregate result is returned...
     
     */
    
    //return...
    id ret = nil;
    
    //open database
    [self openDatabase];
    
    //add data to return array for these columns...
    sqlite3_stmt *statement;
    if(sqlite3_prepare_v2(sqlite3Database, [query UTF8String], -1, &statement, nil) == SQLITE_OK){
        
        if(sqlite3_step(statement) == SQLITE_ROW){
            
            //colData depends on the colType...
            id colData = nil;
            
            //
            BOOL isAggregate = FALSE;
            if([[query uppercaseString] containsString:@"COUNT("]){
                colData = [NSNumber numberWithInt:sqlite3_column_int(statement, 0)];
                isAggregate = TRUE;
            }
            if([[query uppercaseString] containsString:@"AVG("] || [[query uppercaseString] containsString:@"SUM("]){
                colData = [NSNumber numberWithDouble:sqlite3_column_double(statement, 0)];
                isAggregate = TRUE;
            }
            if([[query uppercaseString] containsString:@"MAX("] || [[query uppercaseString] containsString:@"MIN("]){
                colData = [NSNumber numberWithDouble:sqlite3_column_double(statement, 0)];
                isAggregate = TRUE;
            }
            
            //if not aggregate, get the data type for the column...
            if(!isAggregate){
            
                char *theField = (char *) sqlite3_column_text(statement, 0);
                ret = [[NSString alloc] initWithUTF8String:theField];
                
                //column data...
                switch (sqlite3_column_type(statement, 0)){
                    case 1: //integer
                        //colData = [NSString stringWithFormat:@"%i", sqlite3_column_int(statement, 0)];
                        colData = [NSNumber numberWithInt:sqlite3_column_int(statement, 0)];
                        break;
                    case 2: //double, float, real
                        //colData = [NSString stringWithFormat:@"%f", sqlite3_column_double(statement, 0)];
                        colData = [NSNumber numberWithDouble:sqlite3_column_double(statement, 0)];
                        break;
                    case 3: //text
                        colData = [[NSString alloc] initWithUTF8String:(char *) sqlite3_column_text(statement, 0)];
                        break;
                    case 4: //blob
                        colData = [[NSData alloc] initWithBytes:sqlite3_column_blob(statement, 0) length:sqlite3_column_bytes(statement, 0)];
                        break;
                    case 5: //null
                        colData = nil;
                        break;
                }
            
            }//isAggregate...
            
            //if we didn't find any data, we need something to add for the object...
            if(colData == nil){
                colData = @"";
            }
            
            //return is colData we found...
            ret = colData;
            
        }
        
        //clean up statement...
        sqlite3_finalize(statement);
        [self closeDatabase];
        
    }
    
    //return...
    return ret;
    
}



//getOneBlob...
-(NSData *)getOneBlob:(NSString *)query{
    [BT_debugger showIt:self message:[NSString stringWithFormat:@"getOneBlob: \"%@\"", query]];
    
    /*
     Returns an NSData containing one value from one column, such as an image.
     Query example: SELECT column FROM tbl_name WHERE column = '12345';
     */
    
    //return...
    NSData *ret = nil;
    
    //open database
    [self openDatabase];
    
    //add data to return array for these columns...
    sqlite3_stmt *statement;
    if(sqlite3_prepare_v2(sqlite3Database, [query UTF8String], -1, &statement, nil) == SQLITE_OK){
        if(sqlite3_step(statement) == SQLITE_ROW){
            int length = sqlite3_column_bytes(statement, 0);
            ret = [NSData dataWithBytes:sqlite3_column_blob(statement, 0) length:length];
        }
        
        //clean up statement...
        sqlite3_finalize(statement);
        [self closeDatabase];
        
    }
    
    //return...
    return ret;
}

//getSimpleStringListWithQuery...
-(NSMutableArray *)getSimpleStringListWithQuery:(NSString *)query{
    [BT_debugger showIt:self message:[NSString stringWithFormat:@"getSimpleStringListWithQuery: \"%@\"", query]];

    /*
     Returns an NSMutableArray of NSStrings. This is a quickie
     helper method if all you need a list of one column from a table
     */
    
    //return...
    NSMutableArray *ret = [[NSMutableArray alloc] init];
    
    //open database
    [self openDatabase];
    
    //add data to return array for these columns...
    sqlite3_stmt *statement;
    if(sqlite3_prepare_v2(sqlite3Database, [query UTF8String], -1, &statement, nil) == SQLITE_OK){
        int cnt = 0;
        while(sqlite3_step(statement) == SQLITE_ROW){
            
            //ROW ROW DATA IN CONSOLE...
            cnt = (cnt + 1);
            NSString *log = @"";
            
            int colCount = sqlite3_column_count(statement);
            for(int i = 0; i < colCount; i++){
                
                //determined dynamically...
                NSString *colName = [[NSString alloc] initWithUTF8String:(char *) sqlite3_column_name(statement, i)];
                id colData = nil;
                
                log = [log stringByAppendingFormat:@"\n%@: ", colName];
                
                //column data...
                switch (sqlite3_column_type(statement, i)){
                    case 1: //integer
                        colData = [NSString stringWithFormat:@"%i", sqlite3_column_int(statement, i)];
                        log = [log stringByAppendingFormat:@"%@", colData];
                        break;
                    case 2: //double, float, real
                        colData = [NSString stringWithFormat:@"%f", sqlite3_column_double(statement, i)];
                        log = [log stringByAppendingFormat:@"%@", colData];
                        break;
                    case 3: //text
                        colData = [[NSString alloc] initWithUTF8String:(char *) sqlite3_column_text(statement, i)];
                        log = [log stringByAppendingFormat:@"%@", colData];
                        break;
                    case 4: //blob
                        colData = [[NSData alloc] initWithBytes:sqlite3_column_blob(statement, i) length:sqlite3_column_bytes(statement, i)];
                        log = [log stringByAppendingFormat:@"%@", @"blob"];
                        break;
                    case 5: //null
                        colData = nil;
                        log = [log stringByAppendingFormat:@"%@", @""];
                        break;
                }
                
                //no data...
                if(colData == nil){
                    colData = @"";
                }

                //add the key/value for this column..
                [ret addObject:colData];
                
            }//for each column...
            
        }//while...
        
        //clean up statement...
        sqlite3_finalize(statement);
        [self closeDatabase];
        
    }//if prepare...
    
    //return array...
    return ret;
    
}

//getItemWithQuery...
-(NSMutableDictionary *)getItemWithQuery:(NSString *)query{
    [BT_debugger showIt:self message:[NSString stringWithFormat:@"getItemWithQuery: \"%@\"", query]];
    
    /*
     Returns an NSMutableDictionary object.
     The dictionary contains key / values for each colum in the select
     */
    
    //return...
    NSMutableDictionary *ret = [[NSMutableDictionary alloc] init];
    
    //open database
    [self openDatabase];
    
    //add data to return array for these columns...
    sqlite3_stmt *statement;
    if(sqlite3_prepare_v2(sqlite3Database, [query UTF8String], -1, &statement, nil) == SQLITE_OK){
        int cnt = 0;
        while(sqlite3_step(statement) == SQLITE_ROW){
            
            cnt = (cnt + 1);
            NSString *log = @"";
            
            int colCount = sqlite3_column_count(statement);
            for(int i = 0; i < colCount; i++){
                
                //determined dynamically...
                NSString *colName = [[NSString alloc] initWithUTF8String:(char *) sqlite3_column_name(statement, i)];
                id colData = nil;
                
                log = [log stringByAppendingFormat:@"\n%@: ", colName];
                
                //column data...
                switch (sqlite3_column_type(statement, i)){
                    case 1: //integer
                        colData = [NSString stringWithFormat:@"%i", sqlite3_column_int(statement, i)];
                        log = [log stringByAppendingFormat:@"%@", colData];
                        break;
                    case 2: //double, float, real
                        colData = [NSString stringWithFormat:@"%f", sqlite3_column_double(statement, i)];
                        log = [log stringByAppendingFormat:@"%@", colData];
                        break;
                    case 3: //text
                        colData = [[NSString alloc] initWithUTF8String:(char *) sqlite3_column_text(statement, i)];
                        log = [log stringByAppendingFormat:@"%@", colData];
                        break;
                    case 4: //blob
                        colData = [[NSData alloc] initWithBytes:sqlite3_column_blob(statement, i) length:sqlite3_column_bytes(statement, i)];
                        log = [log stringByAppendingFormat:@"%@", @"blob"];
                        break;
                    case 5: //null
                        colData = nil;
                        log = [log stringByAppendingFormat:@"%@", @""];
                        break;
                }
                
                //if we didn't find any data, we need something to add for the object...
                if(colData == nil){
                    colData = @"";
                }
                
                //add the key/value for this column..
                [ret setObject:colData forKey:colName];
                
            }//for each column...
            
            
        }//while...
        
        //clean up statement...
        sqlite3_finalize(statement);
        [self closeDatabase];
        
    }//if prepare...
    
    //return NSDictionary...
    return ret;
}



//getItemsWithQuery...
-(NSMutableArray *)getItemsWithQuery:(NSString *)query{
    [BT_debugger showIt:self message:[NSString stringWithFormat:@"getItemsWithQuery: \"%@\"", query]];
    
    /*
     Returns an NSMutableArray of NSMutableDictionary objects.
     Each object in the list represents one row of data.
     The key for each item is the column name from the table.
     */
    
    //return...
    NSMutableArray *ret = [[NSMutableArray alloc] init];
    
    //open database
    [self openDatabase];
    
    //add data to return array for these columns...
    sqlite3_stmt *statement;
    if(sqlite3_prepare_v2(sqlite3Database, [query UTF8String], -1, &statement, nil) == SQLITE_OK){
        int cnt = 0;
        while(sqlite3_step(statement) == SQLITE_ROW){
            
            cnt = (cnt + 1);
            
            int colCount = sqlite3_column_count(statement);
            NSMutableDictionary *thisRow = [[NSMutableDictionary alloc] init];
            for(int i = 0; i < colCount; i++){
                
                //determined dynamically...
                NSString *colName = [[NSString alloc] initWithUTF8String:(char *) sqlite3_column_name(statement, i)];
                id colData = nil;
                
                //column data...
                switch (sqlite3_column_type(statement, i)){
                    case 1: //integer
                        colData = [NSString stringWithFormat:@"%i", sqlite3_column_int(statement, i)];
                        break;
                    case 2: //double, float, real
                        colData = [NSString stringWithFormat:@"%f", sqlite3_column_double(statement, i)];
                        break;
                    case 3: //text
                        colData = [[NSString alloc] initWithUTF8String:(char *) sqlite3_column_text(statement, i)];
                        break;
                    case 4: //blob
                        colData = [[NSData alloc] initWithBytes:sqlite3_column_blob(statement, i) length:sqlite3_column_bytes(statement, i)];
                        break;
                    case 5: //null
                        colData = nil;
                        break;
                }
                
                //if we didn't find any data, we need something to add for the object...
                if(colData == nil){
                    colData = @"";
                }
                
                //add the key/value for this column..
                [thisRow setObject:colData forKey:colName];
                
            }//for each column...
            
            //add this column to the result...
            [ret addObject:thisRow];
            
        }//while...
        
        //clean up statement...
        sqlite3_finalize(statement);
        [self closeDatabase];
        
    }//if prepare...
    
    //return array...
    return ret;
}



//getBtItemWithId...
-(BT_item *)getBtItemWithId:(NSString *)theId{
    [BT_debugger showIt:self message:[NSString stringWithFormat:@"getBtItemWithId: \"%@\"", theId]];
    
    //return item, assume nil...
    BT_item *ret = nil;
    
    if([theId length] > 0){
    
        NSMutableArray *itemRows = [self getItemsWithQuery:[NSString stringWithFormat:@"SELECT * FROM %@ WHERE guid = '%@'", bt_itemsTableName, theId]];
        if(itemRows.count == 1){
            
            //alloc return item...
            ret = [[BT_item alloc] init];
            
            //item data is dictionary in first row...
            NSMutableDictionary *itemData = [itemRows objectAtIndex:0];
            
            NSString *itemId = ([itemData objectForKey:@"guid"]) ? [itemData objectForKey:@"guid"] : @"n/a";
            NSString *itemType = ([itemData objectForKey:@"itemType"]) ? [itemData objectForKey:@"itemType"] : @"n/a";
            NSString *itemNickname = ([itemData objectForKey:@"nickname"]) ? [itemData objectForKey:@"nickname"] : @"n/a";
            NSString *jsonString = ([itemData objectForKey:@"jsonVars"]) ? [itemData objectForKey:@"jsonVars"] : @"{}";
            NSString *createdUNIX = ([itemData objectForKey:@"createdUNIX"]) ? [itemData objectForKey:@"createdUNIX"] : @"0";
            NSString *modifiedUNIX = ([itemData objectForKey:@"modifiedUNIX"]) ? [itemData objectForKey:@"modifiedUNIX"] : @"0";
            NSString *syncedUNIX = ([itemData objectForKey:@"syncedUNIX"]) ? [itemData objectForKey:@"syncedUNIX"] : @"0";
            
            //jsonVars from data...
            NSError *error;
            NSData *jsonData = nil;
            NSDictionary *jsonDict = nil;
            if([jsonString length] > 4){
                jsonData = [jsonString dataUsingEncoding:NSUTF8StringEncoding];
                jsonDict = [NSJSONSerialization JSONObjectWithData:jsonData options:NSJSONReadingMutableContainers error:&error];
            }
            
            //setup BT_item...
            [ret setItemId:itemId];
            [ret setItemType:itemType];
            [ret setItemNickname:itemNickname];
            
            //add each item from jsonVars string in data...
            NSMutableDictionary *dict = [[NSMutableDictionary alloc] init];
            NSArray *sortedKeys = [[jsonDict allKeys] sortedArrayUsingSelector: @selector(caseInsensitiveCompare:)];
            for(NSString *key in sortedKeys){
                [dict setObject:[jsonDict objectForKey:key] forKey:key];
            }
            
            //make sure dict has itemId, itemType, itemNickname...
            [dict setObject:itemId forKey:@"itemId"];
            [dict setObject:itemType forKey:@"itemType"];
            [dict setObject:itemNickname forKey:@"itemNickname"];

            //make sure dict has createdUNIX, modifiedUNIX, syncedUNIX...
            [dict setObject:createdUNIX forKey:@"createdUNIX"];
            [dict setObject:modifiedUNIX forKey:@"modifiedUNIX"];
            [dict setObject:syncedUNIX forKey:@"syncedUNIX"];
            
            //assign jsonDict...
            [ret setJsonVars:dict];
            
        }else{
            [BT_debugger showIt:self message:[NSString stringWithFormat:@"getBtItemWithId: There are --%lu-- rows in the database with itemId: \"%@\"", (unsigned long)itemRows.count, theId]];
        }
        
    }else{
        [BT_debugger showIt:self message:[NSString stringWithFormat:@"getBtItemWithId: No itemId provided%@", @""]];
    }
    
    //return...
    return ret;

}


//getBtItemsWithQuery...
-(NSMutableArray *)getBtItemsWithQuery:(NSString *)query{
    [BT_debugger showIt:self message:[NSString stringWithFormat:@"getBtItemsWithQuery: \"%@\"", query]];
    
    NSMutableArray *ret = [[NSMutableArray alloc] init];
    NSMutableArray *itemRows = [self getItemsWithQuery:query];
    if(itemRows.count > 0){
        
        int x = 0;
        for(x = 0; x < itemRows.count; x++){
        
            //alloc return item...
            BT_item *thisItem = [[BT_item alloc] init];
        
            //item data is dictionary in first row...
            NSMutableDictionary *itemData = [itemRows objectAtIndex:x];
        
            NSString *itemId = ([itemData objectForKey:@"guid"]) ? [itemData objectForKey:@"guid"] : @"n/a";
            NSString *itemType = ([itemData objectForKey:@"itemType"]) ? [itemData objectForKey:@"itemType"] : @"n/a";
            NSString *itemNickname = ([itemData objectForKey:@"nickname"]) ? [itemData objectForKey:@"nickname"] : @"n/a";
            NSString *jsonString = ([itemData objectForKey:@"jsonVars"]) ? [itemData objectForKey:@"jsonVars"] : @"{}";
            NSString *createdUNIX = ([itemData objectForKey:@"createdUNIX"]) ? [itemData objectForKey:@"createdUNIX"] : @"0";
            NSString *modifiedUNIX = ([itemData objectForKey:@"modifiedUNIX"]) ? [itemData objectForKey:@"modifiedUNIX"] : @"0";
            NSString *syncedUNIX = ([itemData objectForKey:@"syncedUNIX"]) ? [itemData objectForKey:@"syncedUNIX"] : @"0";
        
            //jsonVars from data...
            NSError *error;
            NSData *jsonData = nil;
            NSDictionary *jsonDict = nil;
            if([jsonString length] > 4){
                jsonData = [jsonString dataUsingEncoding:NSUTF8StringEncoding];
                jsonDict = [NSJSONSerialization JSONObjectWithData:jsonData options:NSJSONReadingMutableContainers error:&error];
            }
        
            //setup BT_item...
            [thisItem setItemId:itemId];
            [thisItem setItemType:itemType];
            [thisItem setItemNickname:itemNickname];
        
            //add each item from jsonVars string in data...
            NSMutableDictionary *dict = [[NSMutableDictionary alloc] init];
            NSArray *sortedKeys = [[jsonDict allKeys] sortedArrayUsingSelector: @selector(caseInsensitiveCompare:)];
            for(NSString *key in sortedKeys){
                [dict setObject:[jsonDict objectForKey:key] forKey:key];
            }
        
            //make sure dict has itemId, itemType, itemNickname...
            [dict setObject:itemId forKey:@"itemId"];
            [dict setObject:itemType forKey:@"itemType"];
            [dict setObject:itemNickname forKey:@"itemNickname"];
        
            //make sure dict has createdUNIX, modifiedUNIX, syncedUNIX...
            [dict setObject:createdUNIX forKey:@"createdUNIX"];
            [dict setObject:modifiedUNIX forKey:@"modifiedUNIX"];
            [dict setObject:syncedUNIX forKey:@"syncedUNIX"];
        
            //assign jsonDict...
            [thisItem setJsonVars:dict];
        
            //add to return array...
            [ret addObject:thisItem];
        
        }
        
    }
    
    //return...
    return ret;
    
}



//updateConfigDataInDatabase...
-(void)updateConfigDataInDatabase:(NSString *)appDataJSONString{
    [BT_debugger showIt:self message:[NSString stringWithFormat:@"updateConfigDataInDatabase: parsing config data%@", @""]];
    
    @try{
        
        //date now so inserts / updates in this batch match...
        NSString *dtNow = [BT_dates getUnixNowUTC];
        
        //before update...
        NSArray *tmpItems = [self getItemsWithQuery:[NSString stringWithFormat:@"SELECT * FROM %@", bt_itemsTableName]];
        [BT_debugger showIt:self message:[NSString stringWithFormat:@"updateConfigDataInDatabase: Database contains %i rows BEFORE update.", (int)[tmpItems count]]];
        
        //Remove previously created tabs everytime. BT_tabs get new Id's in the control panel constantly...
        [self executeQuery:[NSString stringWithFormat:@"DELETE FROM %@ WHERE itemTypeLabel = 'Tab'", bt_itemsTableName]];
        
        
        //updateItems, insertItems, deleteItems...
        NSMutableArray *updateItems = [[NSMutableArray alloc] init];
        NSMutableArray *insertItems = [[NSMutableArray alloc] init];
        
        //load up existing guids (we test against this list before inserting or updating...
        NSMutableArray *existingGuids = [self getSimpleStringListWithQuery:[NSString stringWithFormat:@"SELECT guid FROM %@", bt_itemsTableName]];
        
        //create dictionary from the JSON string...
        NSError *jsonError = nil;
        NSData *jsonDataFromString = [appDataJSONString dataUsingEncoding:NSUTF8StringEncoding];
        NSDictionary *jsonParser = [NSJSONSerialization JSONObjectWithData:jsonDataFromString options:0 error:&jsonError];
        
        if(jsonError != nil){
            [BT_debugger showIt:self message:[NSString stringWithFormat:@"ERROR parsing JSON in updateConfigDataInDatabase: %@", jsonError.description]];
        }else{
            
            //get the first item in the list of BT_item, it should be a BT_item.itemType == "BT_app"
            if([[jsonParser objectForKey:@"BT_appConfig"] objectForKey:@"BT_items"]){
                NSArray *tmpItems = [[jsonParser objectForKey:@"BT_appConfig"] objectForKey:@"BT_items"];
                if([tmpItems count] < 1){
                    [BT_debugger showIt:self message:[NSString stringWithFormat:@"the BT_items array is empty?%@", @""]];
                }else{
                    NSDictionary *thisApp = [tmpItems objectAtIndex:0];
                    if([[thisApp objectForKey:@"itemType"] isEqualToString:@"BT_app"]){
                        
                        //properties to find and fill from JSON...
                        NSMutableDictionary *props = [[NSMutableDictionary alloc] init];
                        [props setObject:@"" forKey:@"buzztouchAppId"];
                        [props setObject:@"" forKey:@"buzztouchAPIKey"];
                        [props setObject:@"" forKey:@"themeItemId"];
                        [props setObject:@"" forKey:@"homeScreenItemId"];
                        [props setObject:@"" forKey:@"splashScreenItemId"];
                        [props setObject:@"" forKey:@"registrationScreenItemId"];
                        [props setObject:@"" forKey:@"registrationRequired"];
                        [props setObject:@"" forKey:@"dataURL"];
                        [props setObject:@"" forKey:@"reportToCloudURL"];
                        [props setObject:@"" forKey:@"registerForPushURL"];
                        [props setObject:@"" forKey:@"allowRotation"];
                        [props setObject:@"" forKey:@"gpsUpdateIntervalSeconds"];
                        [props setObject:@"" forKey:@"promptForPushNotifications"];
                        [props setObject:@"" forKey:@"currentMode"];
                        
                        //find / fill each key...
                        NSMutableDictionary *propsDict = [[NSMutableDictionary alloc] init];
                        for(id key in props){
                            if([thisApp objectForKey:key]){
                                NSString *val = [thisApp objectForKey:key];
                                [propsDict setObject:val forKey:key];
                            }
                        }
                        
                        //appCoreProperties to insert or update in local SqlLite database...
                        NSMutableDictionary *appCoreProperties = [[NSMutableDictionary alloc] init];
                        [appCoreProperties setObject:@"appCoreProperties" forKey:@"guid"];
                        [appCoreProperties setObject:@"appCoreProperties" forKey:@"itemType"];
                        [appCoreProperties setObject:@"appCoreProperties" forKey:@"nickname"];
                        [appCoreProperties setObject:@"" forKey:@"parentItemGuid"];
                        
                        NSString *jsonString = @"{}";
                        
                        //jsonVars from data...
                        NSError *error;
                        NSData *jsonData = [NSJSONSerialization dataWithJSONObject:propsDict options:0 error:&error];
                        if(jsonData){
                            jsonString = [[NSString alloc] initWithBytes:[jsonData bytes] length:[jsonData length] encoding:NSUTF8StringEncoding];
                        }
                        
                        [appCoreProperties setObject:jsonString forKey:@"jsonVars"];
                        [appCoreProperties setObject:@"0" forKey:@"hasChildItems"];
                        [appCoreProperties setObject:@"0" forKey:@"orderIndex"];
                        [appCoreProperties setObject:dtNow forKey:@"modifiedUNIX"];
                        [appCoreProperties setObject:@"appCoreProperties" forKey:@"itemTypeLabel"];
                        [appCoreProperties setObject:@"appCoreProperties" forKey:@"controlPanelItemType"];

                        //insert or update appCoreProperties...
                        if([existingGuids containsObject:@"appCoreProperties"]){
                            [self updateRowInTable:bt_itemsTableName withRow:appCoreProperties];
                        }else{
                            [appCoreProperties setObject:dtNow forKey:@"createdUNIX"];
                            [self insertRowIntoTable:bt_itemsTableName withRow:appCoreProperties];
                        }
                        
                        //process each item type from the JSON config file...
                        NSArray *itemTypes = [NSArray arrayWithObjects:@"BT_themes", @"BT_tabs", @"BT_menus", @"BT_screens", @"BT_files", @"BT_images", nil];
                        NSArray *itemTypeLabels = [NSArray arrayWithObjects:@"Theme", @"Tab", @"Menu", @"Screen", @"File", @"Image", nil];
                        for(int x = 0; x < itemTypes.count; x++) {
                            
                            NSString *thisList = [itemTypes objectAtIndex:x];
                            if([thisApp objectForKey:thisList]){
                                
                                NSArray *tmpItems = [thisApp objectForKey:thisList];
                                for(NSDictionary *json in tmpItems){
                                    
                                    //row vars, must have itemId...
                                    NSString *itemId = ([json objectForKey:@"itemId"]) ? [json objectForKey:@"itemId"] : @"";
                                    if([itemId length] > 0){
                                        
                                        NSString *itemType = ([json objectForKey:@"itemType"]) ? [json objectForKey:@"itemType"] : @"";
                                        NSString *nickname = ([json objectForKey:@"itemNickname"]) ? [json objectForKey:@"itemNickname"] : @"";
                                        NSString *parentItemGuid = ([json objectForKey:@"parentItemGuid"]) ? [json objectForKey:@"parentItemGuid"] : @"";
                                        NSString *jsonString = @"{}";
                                        
                                        //jsonVars from data...
                                        NSError *error;
                                        NSData *jsonData = [NSJSONSerialization dataWithJSONObject:json options:0 error:&error];
                                        if(jsonData){
                                            jsonString = [[NSString alloc] initWithBytes:[jsonData bytes] length:[jsonData length] encoding:NSUTF8StringEncoding];
                                        }
                                        
                                        //one row...
                                        NSMutableDictionary *row = [[NSMutableDictionary alloc] init];
                                        [row setObject:itemId forKey:@"guid"];
                                        [row setObject:itemType forKey:@"itemType"];
                                        [row setObject:nickname forKey:@"nickname"];
                                        [row setObject:parentItemGuid forKey:@"parentItemGuid"];
                                        [row setObject:jsonString forKey:@"jsonVars"];
                                        [row setObject:@"0" forKey:@"hasChildItems"];
                                        [row setObject:@"0" forKey:@"orderIndex"];
                                        [row setObject:dtNow forKey:@"modifiedUNIX"];
                                        [row setObject:[itemTypeLabels objectAtIndex:x] forKey:@"itemTypeLabel"];
                                        [row setObject:[itemTypeLabels objectAtIndex:x] forKey:@"controlPanelItemType"];
                                        
                                        //insert or update...
                                        if([existingGuids containsObject:itemId]){
                                            [updateItems addObject:row];
                                        }else{
                                            [row setObject:dtNow forKey:@"createdUNIX"];
                                            [insertItems addObject:row];
                                        }
                                        
                                    }//itemId...
                                }//for tmpItems...
                            }//json has this item...
                        }//for each item in the list...
                        
                    }//BT_app...
                }
            }else{
                [BT_debugger showIt:self message:[NSString stringWithFormat:@"there are no BT_items in the configuration data?%@", @""]];
            }
            
            //updateItems...
            if([updateItems count] > 0){
                [self updateRowsInTable:bt_itemsTableName withRows:updateItems];
            }
            
            //insertItems...
            if([insertItems count] > 0){
                [self insertRowsIntoTable:bt_itemsTableName withRows:insertItems];
            }
            
            //after update...
            NSArray *tmpItems = [self getItemsWithQuery:[NSString stringWithFormat:@"SELECT * FROM %@", bt_itemsTableName]];
            [BT_debugger showIt:self message:[NSString stringWithFormat:@"updateConfigDataInDatabase: Database contains %i rows AFTER update.", (int)[tmpItems count]]];
            
        }
        
    }@catch (NSException * e){
        [BT_debugger showIt:self message:[NSString stringWithFormat:@"error parsing application data in updateConfigDataInDatabase: %@", e.description]];
    }
    
    
    
}



@end

































