/*
 *	Copyright, Buzztouch
 *
 *	All rights reserved.
 *
 *	Redistribution and use in source and binary forms, with or without modification, are
 *	permitted provided that the following conditions are met:
 *
 *	Redistributions of source code must retain the above copyright notice which includes the
 *	name(s) of the copyright holders. It must also retain this list of conditions and the
 *	following disclaimer.
 *
 *	Redistributions in binary form must reproduce the above copyright notice, this list
 *	of conditions and the following disclaimer in the documentation and/or other materials
 *	provided with the distribution.
 *
 *	The name Buzztouch (or buzztouch) nor the names of its contributors
 *	may be used to endorse or promote products derived from this software without specific
 *	prior written permission.
 *
 *	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 *	ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 *	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 *	IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 *	INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *	NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *	PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 *	WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *	ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
 *	OF SUCH DAMAGE.
 */

#import <MobileCoreServices/UTCoreTypes.h>
#import <CoreLocation/CoreLocation.h>
#import <mach/mach.h>
#import <MessageUI/MessageUI.h>
#import "BT_strings.h"
#import "BT_debugger.h"
#import "BT_reachability.h"
#import "BT_device.h"
#import <CoreText/CoreText.h>


@implementation BT_device

-(id)init{
    if((self = [super init])){
        [BT_debugger showIt:self message:[NSString stringWithFormat:@"INIT"]];
        
        //current device info...
        UIDevice* device = [UIDevice currentDevice];
        
        //current locale info...
        NSLocale *theLocale = [NSLocale currentLocale];
        NSString *currencySymbol = [theLocale objectForKey:NSLocaleCurrencySymbol];
        NSString *currencyCode = [theLocale objectForKey:NSLocaleCurrencyCode];
        NSString *countryCode = [[NSLocale currentLocale] objectForKey: NSLocaleCountryCode];
        NSString *languageCode = [[NSLocale preferredLanguages] objectAtIndex:0];
        
        //we only want the "main" part of the language, not the local (en not en-US)...
        if([languageCode length] > 2){
            languageCode = [languageCode substringToIndex:2];
        }
        
        //deviceType...
        [self setDeviceType:@"ios"];
        
        //create a unique-id if we have not saved one yet.
        NSString *tmpUUID = [BT_strings getOrCreateDeviceId];
        [self setDeviceId:tmpUUID];
        
        [self setDeviceName:[device name]];
        [self setDeviceState:@"foreground"];
        [self setDeviceModel:[device model]];
        [self setDeviceSystem:[device systemName]];
        [self setDeviceVersion:[device systemVersion]];
        [self setDeviceMemory:[self getCurrentMemoryUsage]];
        [self setDeviceWidth:[UIScreen mainScreen].bounds.size.width];
        [self setDeviceHeight:[UIScreen mainScreen].bounds.size.height];
        [self setDeviceLastLatitude:@"0"];
        [self setDeviceLastLongitude:@"0"];
        [self setDeviceLastLocationReport:@"0"];
        [self setDeviceLastReportToCloud:@"0"];
        [self setDeviceConnectionType:@"none"];
        [self setIsOnline:FALSE];
        [self setCanTakePictures:FALSE];
        [self setCanTakeVideos:FALSE];
        [self setCanMakePhoneCalls:FALSE];
        [self setCanSendEmails:FALSE];
        [self setCanSendSMS:FALSE];
        [self setCanReportLocation:FALSE];
        [self setDeviceCountryCode:countryCode];
        [self setDeviceLangCode:languageCode];
        [self setDeviceCurrencyCode:currencyCode];
        [self setDeviceCurrencySymbol:currencySymbol];
        [self setDeviceSpeed:-1];
        [self setDeviceHeading:-1];

        
        //if this is an iPad, flag it...
        if([[UIDevice currentDevice] userInterfaceIdiom] == UIUserInterfaceIdiomPad) {
            [self setIsIPad:TRUE];
        }else{
            [self setIsIPad:FALSE];
        }
        
        //can this device make phone calls?
        NSRange modelRange = [[device model] rangeOfString:@"iPhone"];
        if(modelRange.location == NSNotFound){
            [self setCanMakePhoneCalls:FALSE];
        }else{
            [self setCanMakePhoneCalls:TRUE];
        }
        
        //can send emails? Just because device is capable doesn't mean mail's been configured.
        Class mailClass = (NSClassFromString(@"MFMailComposeViewController"));
        if(mailClass != nil){
            if([mailClass canSendMail]){
                [self setCanSendEmails:TRUE];
            }
        }
        
        //can send SMS? Just because device is capable doesn't mean it can send.
        Class SMSClass = (NSClassFromString(@"MFMessageComposeViewController"));
        if(SMSClass != nil){
            if([SMSClass canSendText]){
                [self setCanSendSMS:TRUE];
            }
        }
        
        /*
         can this device report it's location? If user has turned off services for this app only, this
         method will return true. In this case, we need to use the locationManager.didFailWithError
         method in the delgate class where we are monitoring the location.
         */
        SEL selector = NSSelectorFromString(@"locationServicesEnabled");
        if([CLLocationManager respondsToSelector:selector]) {
            [self setCanReportLocation:[CLLocationManager locationServicesEnabled]];
        }
        
        //camera and or video support?
        if([UIImagePickerController isSourceTypeAvailable:UIImagePickerControllerSourceTypeCamera]){
            NSArray *media = [UIImagePickerController availableMediaTypesForSourceType:UIImagePickerControllerSourceTypeCamera];
            if([media containsObject:(id)kUTTypeImage]){
                [self setCanTakePictures:TRUE];
            }
            if([media containsObject:(id)kUTTypeMovie]){
                [self setCanTakeVideos:TRUE];
            }
        }
        
        //show all the custom font names listed in BT_info.plist...
        [self setCustomFontNames:[[NSMutableArray alloc] init]];
        NSDictionary* infoDict = [[NSBundle mainBundle] infoDictionary];
        NSArray *fontNames = [infoDict objectForKey:@"UIAppFonts"];
        if([fontNames count] > 1){
            for (NSString *fontName in fontNames){
                
                NSURL *url = [[NSBundle mainBundle] URLForResource:fontName withExtension:NULL];
                NSData *fontData = [NSData dataWithContentsOfURL:url];
                CGDataProviderRef fontDataProvider = CGDataProviderCreateWithCFData((__bridge CFDataRef)fontData);
                
                //must call this line BEFORE registering font...
                [UIFont familyNames];
                
                //load the custom font...
                CGFontRef customFont = CGFontCreateWithDataProvider(fontDataProvider);
                CFErrorRef error;
                CTFontManagerRegisterGraphicsFont(customFont, &error);
                NSString *customFontName = CFBridgingRelease(CGFontCopyFullName(customFont));
                CGFontRelease(customFont);
                CGDataProviderRelease(fontDataProvider);
                
                [self.customFontNames addObject:customFontName];
                
            }
        }
        
    }
    return self;
}

//register for push notifications...
-(void)registerForPushNotifications:(NSString *)registerURL{
    [BT_debugger showIt:self theMessage:[NSString stringWithFormat:@"registerForPushNotifications: %@", @""]];
    
    //process the URL request...
    NSError* error;
    NSURL *useURL = [NSURL URLWithString:registerURL];
    NSString *response = [NSString stringWithContentsOfURL:useURL encoding:NSASCIIStringEncoding error:&error];
    [BT_debugger showIt:self theMessage:[NSString stringWithFormat:@"registerForPushNotifications: Response: %@", response]];
    
}

//unRegister for push notifications...
-(void)unRegisterForPushNotifications:(NSString *)registerURL{
    [BT_debugger showIt:self theMessage:[NSString stringWithFormat:@"unRegisterForPushNotifications: %@", @""]];
    
    
    //process the URL request...
    NSError* error;
    NSURL *useURL = [NSURL URLWithString:registerURL];
    NSString *response = [NSString stringWithContentsOfURL:useURL encoding:NSASCIIStringEncoding error:&error];
    [BT_debugger showIt:self theMessage:[NSString stringWithFormat:@"unRegisterForPushNotifications: Response: %@", response]];
    
    
}

//getInfoForDevice...
-(NSMutableDictionary *)getInfoForDevice{
    [BT_debugger showIt:self theMessage:[NSString stringWithFormat:@"getInfoForDevice%@", @""]];
    
    //create a dictionary to return...
    NSMutableDictionary *ret = [[NSMutableDictionary alloc] init];
    
    //add each property and value...
    [ret setObject:@"ios" forKey:@"deviceType"];
    [ret setObject:[self deviceId] forKey:@"deviceId"];
    [ret setObject:[self deviceModel] forKey:@"deviceModel"];
    [ret setObject:[self deviceName] forKey:@"deviceName"];
    [ret setObject:[self deviceState] forKey:@"deviceState"];
    [ret setObject:[self deviceSystem] forKey:@"deviceSystem"];
    [ret setObject:[self deviceMemory] forKey:@"deviceMemory"];
    [ret setObject:[self deviceVersion] forKey:@"deviceVersion"];
    [ret setObject:[self deviceLastLatitude] forKey:@"deviceLastLatitude"];
    [ret setObject:[self deviceLastLongitude] forKey:@"deviceLastLongitude"];
    [ret setObject:[self deviceLastLocationReport] forKey:@"deviceLastLocationReport"];
    [ret setObject:[self deviceLastReportToCloud] forKey:@"deviceLastReportToCloud"];
    [ret setObject:[NSString stringWithFormat:@"%i", [self deviceSpeed]] forKey:@"deviceSpeed"];
    [ret setObject:[NSString stringWithFormat:@"%i", [self deviceHeading]] forKey:@"deviceHeading"];
    [ret setObject:[self deviceConnectionType] forKey:@"deviceConnectionType"];
    [ret setObject:[self deviceCountryCode] forKey:@"deviceCountryCode"];
    [ret setObject:[self deviceLangCode] forKey:@"deviceLangCode"];
    [ret setObject:[self deviceCurrencyCode] forKey:@"deviceCurrencyCode"];
    [ret setObject:[self deviceCurrencySymbol] forKey:@"deviceCurrencySymbol"];
    [ret setObject:[NSString stringWithFormat:@"%i", [self deviceWidth]] forKey:@"deviceWidth"];
    [ret setObject:[NSString stringWithFormat:@"%i", [self deviceHeight]] forKey:@"deviceHeight"];
    [ret setObject:[NSString stringWithFormat:@"%i", [self isIPad]] forKey:@"isIPad"];
    [ret setObject:[NSString stringWithFormat:@"%i", [self isOnline]] forKey:@"isOnline"];
    [ret setObject:[NSString stringWithFormat:@"%i", [self canReportLocation]] forKey:@"canReportLocation"];
    [ret setObject:[NSString stringWithFormat:@"%i", [self canTakePictures]] forKey:@"canTakePictures"];
    [ret setObject:[NSString stringWithFormat:@"%i", [self canTakeVideos]] forKey:@"canTakeVideos"];
    [ret setObject:[NSString stringWithFormat:@"%i", [self canMakePhoneCalls]] forKey:@"canMakePhoneCalls"];
    [ret setObject:[NSString stringWithFormat:@"%i", [self canSendEmails]] forKey:@"canSendEmails"];
    [ret setObject:[NSString stringWithFormat:@"%i", [self canSendSMS]] forKey:@"canSendSMS"];
    
    //return...
    return ret;
    
}

//printInfoForDevice...
-(void)printInfoForDevice{
    [BT_debugger showIt:self theMessage:[NSString stringWithFormat:@"printInfoForDevice%@", @""]];
    
    //list device info (in alphabeticaly order by key name)...
    NSMutableDictionary *deviceInfo = [self getInfoForDevice];
    NSArray *sortedKeys = [[deviceInfo allKeys] sortedArrayUsingSelector: @selector(caseInsensitiveCompare:)];
    for(NSString *key in sortedKeys) {
        [BT_debugger showIt:self message:[NSString stringWithFormat:@"Device: %@ = %@", key, [deviceInfo objectForKey:key]]];
    }
    
    //list custom font names...
    if([self.customFontNames count] > 0){
        NSString *fontNames = @"";
        for(NSString *fontName in self.customFontNames){
            fontNames = [fontNames stringByAppendingString:[NSString stringWithFormat:@"\"%@\", ", fontName]];
        }
        //remove trailing comma and space...
        if([fontNames length] > 5){
            fontNames = [fontNames substringToIndex:[fontNames length] -2];
        }
        [BT_debugger showIt:self message:[NSString stringWithFormat:@"Device: Custom UIAppFonts included in .plist %@", fontNames]];
    }else{
        [BT_debugger showIt:self message:[NSString stringWithFormat:@"Device: No custom UIAppFonts included in .plist%@", @""]];
    }
    

}




//getCurrentMemoryUsage...
-(NSString *)getCurrentMemoryUsage{
    [BT_debugger showIt:self theMessage:[NSString stringWithFormat:@"getCurrentMemoryUsage%@", @""]];

    /*
        Note: the memory value return from this method will never match what Xcode (Instruments, 
        memory graph, etc) will show. There really is no way to compare the two. Memory mapping
        on iOS is sophisticated and it's way beyond these notes to explain. 
     
        Use this routine to get the memory at a given time, get it again and compare, it's all relative.
     
    */
    
    
    NSString *ret = @"";
    struct task_basic_info info;
    mach_msg_type_number_t size = sizeof(info);
    kern_return_t kerr = task_info(mach_task_self(), TASK_BASIC_INFO, (task_info_t)&info, &size);
    if(kerr == KERN_SUCCESS){
        ret = [NSByteCountFormatter stringFromByteCount:info.resident_size countStyle:NSByteCountFormatterCountStyleMemory];
    }else{
        ret = [NSString stringWithFormat:@"%s", mach_error_string(kerr)];
    }
    
    //remember...
    [self setDeviceMemory:ret];

    //return...
    return ret;
    
}




@end







